from typing import Any, Generic
from typing import TypeVar

from strawberry.types.fields.resolver import StrawberryResolver

T = TypeVar('T')


class FieldResolver(Generic[T], StrawberryResolver[T]):
    def __call__(self, *args: str, **kwargs: Any) -> T:
        if not callable(self.wrapped_func):
            raise Exception(
                f"Attempted to call resolver {self.wrapped_func} with not callable function "
                f"{self.wrapped_func}"
            )
        if len(args) > 0 and hasattr(self.wrapped_func, 'metadata__'):
            args = (getattr(self.wrapped_func, 'metadata__'),)
        return self.wrapped_func(*args, **kwargs)



import asyncio
import inspect
import logging
import traceback
from typing import Any, Annotated

from .utils import Utils
from ..common.enum.platform import PlatFormType
from ..core.module.compiler import ModuleCompiler
from ..core.platform.platform import PlatformAdapter
from ..core.platform.platform_litestar import PlatformLitestar
from ..core.platform.platform_fastapi import PlatformFastAPI

Injected = Annotated[int, 'Test']

FORMAT = '%(levelname)-10s%(asctime)-15s - %(message)s'
DATE_FORMAT = '%b %d %H:%M:%S'

logging.basicConfig(format=FORMAT, datefmt=DATE_FORMAT)


class AppNestipyContext:
    server_kwargs: dict
    adapter: PlatformAdapter
    compiler: ModuleCompiler
    app: Any
    message: Any = None
    logger: logging.Logger

    def __init__(self, module, platform=PlatFormType, **kwargs):
        self.platform = platform
        self.compiler = ModuleCompiler(module)
        self.adapter = PlatformFastAPI() if platform == PlatFormType.FASTAPI else PlatformLitestar()
        self.kwargs = kwargs
        self.config_logger()

    def config_logger(self):
        formatter = logging.Formatter(fmt=FORMAT, datefmt=DATE_FORMAT)
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(handler)

    async def __call__(self, scope, receive, send, **kwargs):
        if scope['type'] == 'lifespan':
            self.message = await self.handle_lifespan(receive)
            await self.app(scope, self.receive, send)
        else:
            # call middleware before
            response = await self.call_middleware(scope, receive, send)
            if not response:
                if hasattr(self, 'app') and self.app is not None:
                    await self.app(scope, receive, send)
                else:
                    raise Exception('Platform Handler not initialized')

    async def receive(self):
        await asyncio.sleep(0.001)
        return self.message

    async def handle_lifespan(self, receive):
        message = await receive()
        if message['type'] == 'lifespan.startup':
            await self.setup()
            await self.on_startup()
        elif message['type'] == 'lifespan.shutdown':
            await self.on_shutdown()
        return message

    async def setup(self):
        try:
            compiled_module = await self.compiler.compile()
            await self.adapter.setup(compiled_module)
            self.app = self.adapter.create_server(**self.kwargs)
            self.logger.info("Application initialized successfully.")
        except Exception as e:
            tb = traceback.format_exc()
            logging.error(e)
            logging.error(tb)
            raise e

    async def call_hooks(self, key):

        if key in self.compiler.hooks.keys():
            hooks = self.compiler.hooks.get(key)
            for hook in hooks:
                try:
                    if inspect.iscoroutinefunction(hook):
                        await hook()
                    else:
                        hook()
                except Exception as e:
                    tb = traceback.format_exc()
                    logging.error(e)
                    logging.error(tb)
                    raise e

    async def call_middleware(self, scope, receive, send):
        middlewares = self.compiler.middlewares
        for middleware in middlewares:
            path = scope["path"]
            match = Utils.match_route(middleware.path, path)
            if match:
                try:
                    if inspect.iscoroutinefunction(middleware.middleware):
                        response = await middleware.middleware(scope, receive, send)
                        if response:
                            return response
                    else:
                        response = middleware.middleware(scope, receive, send)
                        if response:
                            return response
                except Exception as e:
                    tb = traceback.format_exc()
                    logging.error(e)
                    logging.error(tb)
                    raise e

    async def on_startup(self):
        await self.call_hooks('on_startup')

    async def on_shutdown(self):
        await self.call_hooks('on_shutdown')



import inspect
from dataclasses import dataclass
from typing import Callable, Any

from pyee import EventEmitter

from .middleware import MiddlewareConsumer
from .provider import ModuleProvider
from ..ioc import NestipyContainer
from ..utils import Utils
from ...common.decorator.injectable import Injectable


@dataclass
class MiddlewareDict:
    path: str
    middleware: Callable


class ModuleCompiler:
    module: Any
    middlewares: list[MiddlewareDict]

    def __init__(self, module: Callable):
        self.module = module
        self.container = NestipyContainer()
        self.module_resolved = {}
        self.hooks = {}
        self.middlewares = []
        self.ee = EventEmitter()
        self.middleware_consumer = MiddlewareConsumer(self)

    def extract_hook_in_module(self, hook, module):
        if hasattr(module, hook) and inspect.ismethod(getattr(module, hook)):
            hook__ = getattr(module, hook)
            hooks = []
            if hook in self.hooks.keys():
                hooks = self.hooks[hook]
            hooks.append(hook__)
            self.hooks[hook] = hooks

    def extract_hooks_in_module(self, module):
        self.extract_hook_in_module("on_startup", module)
        self.extract_hook_in_module("on_shutdown", module)

    @classmethod
    def update_module_import(cls, module):
        setattr(module, 'imports', [im for im in (module.imports or []) if im is not None])

    async def register_and_resolve_provider(self, module, init=False):
        self.update_module_import(module)
        await self.resolve_providers_of_module(module)
        imports = getattr(module, 'imports')
        global_module = []
        global_providers = []
        if init:
            ins_m = module()
            self.module_resolved[module.__name__] = ins_m
            self.extract_hooks_in_module(ins_m)
            global_providers = self.extract_provider(module)
            old_global_module = [m for m in imports if
                                 hasattr(m, "is_global__") and m.is_global__]
            for index, gb_module in enumerate(old_global_module):
                # set global providers for  all
                setattr(gb_module, 'providers', global_providers + getattr(gb_module, 'providers'))
                self.modify_module_imports(gb_module, old_global_module[0:index + 1])
                await self.register_and_resolve_provider(gb_module)
                global_module.append(gb_module)

        for module_import in [m for m in imports if
                              m not in global_module and m.__name__ not in self.module_resolved.keys()]:
            # set global providers for  all
            setattr(module_import, 'providers', global_providers + getattr(module_import, 'providers'))
            self.modify_module_imports(module_import, global_module)
            await self.register_and_resolve_provider(module_import)
            instance = module_import()
            self.module_resolved[module_import.__name__] = instance
            self.extract_hooks_in_module(instance)

    def modify_module_imports(self, module, global_modules):
        self.update_module_import(module)
        filtered_global_modules = [m for m in global_modules if m != module]
        module_imports: list = (module.imports or [])
        module_imports_with_global = filtered_global_modules + module_imports
        setattr(module, 'imports', module_imports_with_global)
        setattr(module, 'container__', self.container)

    async def recreate_async_provider(self, module):
        if hasattr(module, 'module_provider__'):
            provider: ModuleProvider = getattr(module, 'module_provider__')
            if provider.provide:
                if provider.use_value is not None:
                    setattr(module, "providers",
                            [Injectable(provider.provide)(lambda: provider.use_value)] + module.providers)
                    return
                if provider.use_factory:
                    inject = self.filter_async_provider_inject(module.imports, provider.inject)
                    value = await self.resolve_use_factory(use_factory=provider.use_factory, inject=inject)
                    setattr(module, "providers", [Injectable(provider.provide)(lambda: value)] + module.providers)
                    return

    async def recreate_module_provider(self, module):
        providers = getattr(module, 'providers')
        instance_of_module_provider = [p for p in providers if isinstance(p, ModuleProvider)]
        module_provider_injectable = self.extract_provider(module)
        for p in instance_of_module_provider:
            if p.provide is None:
                pass
            if p.use_value is not None:
                module_provider_injectable = [Injectable(p.provide)(lambda: p.use_value)] + module_provider_injectable
                continue
            if p.use_factory is not None:
                value = await self.resolve_use_factory(use_factory=p.use_factory, inject=module_provider_injectable)
                module_provider_injectable = [Injectable(p.provide)(lambda: value)] + module_provider_injectable
                continue
        setattr(module, 'providers', module_provider_injectable)

    @classmethod
    def extract_provider(cls, module):
        return [p for p in module.providers if hasattr(p, 'injectable__')]

    def filter_async_provider_inject(self, modules: list, inject: list):
        providers = []
        for module in modules:
            providers = providers + self.extract_provider(module)
        return [p for p in inject if p in providers]

    async def resolve_use_factory(self, use_factory, inject=None):
        if inject is None:
            inject = []
        factory_signature = inspect.signature(use_factory).parameters
        factory_dependencies = {}
        for param_name, param in factory_signature.items():
            if param.annotation in inject:
                dep_value = self.container.instances.get(param.annotation)
                factory_dependencies[param_name] = dep_value
        if inspect.iscoroutinefunction(use_factory):
            return await use_factory(**factory_dependencies)
        return use_factory(**factory_dependencies)

    async def resolve_providers_of_module(self, module):
        await self.recreate_async_provider(module)
        await self.recreate_module_provider(module)
        providers = self.extract_provider(module)
        for p in providers:
            is_middleware = hasattr(p, 'middleware__')
            if inspect.isclass(p):
                instance = self.container.resolve(p, module)
                self.put_module_provider_instance(module, p, instance, is_middleware=is_middleware)
            else:
                token = p.token__
                instance = self.container.resolve_method(p, token=token)
                self.put_module_provider_instance(module, token, instance, is_middleware=is_middleware)

        await self.resolve_controllers_of_module(module)
        self.extract_middleware_of_module(module)

    async def resolve_controllers_of_module(self, module):
        controllers = [ctrl for ctrl in module.controllers if hasattr(ctrl, 'controller__')]
        for ctrl in controllers:
            if inspect.isclass(ctrl):
                self.container.resolve(ctrl, module)
            else:
                token = ctrl.token__
                self.container.resolve_method(ctrl, token=token)

    @classmethod
    def put_module_provider_instance(cls, module, token, instance, is_middleware=False):
        if is_middleware:
            cls.put_module_element(module, 'middleware_instances__', token, instance)
        else:
            cls.put_module_element(module, 'provider_instances__', token, instance)

    @classmethod
    def put_module_element(cls, module, element_type, key, value):
        instances__ = {}
        if hasattr(module, element_type):
            instances__ = getattr(module, element_type)
        instances__[key] = value
        setattr(module, element_type, instances__)

    # MIDDLEWARE
    @classmethod
    def get_middleware_of_handler(cls, handler, path='path__'):
        path = getattr(handler, path) or ''
        if hasattr(handler, 'middleware__'):
            return path, getattr(handler, 'middleware__') or []
        return path, []

    def apply_middleware_to_path(self, module, path, middlewares: list):
        transformed_middleware = []
        for m in middlewares:
            if inspect.isclass(m):
                instance = self.container.resolve(m, module)
                middleware = getattr(instance, 'use')
                transformed_middleware.append(MiddlewareDict(path=path, middleware=middleware))
            elif inspect.isfunction(m) or inspect.ismethod(m):
                transformed_middleware.append(MiddlewareDict(path=path, middleware=m))

        self.middlewares += transformed_middleware
        return transformed_middleware

    def apply_middleware_to_ctrl(self, module, ctrl, middlewares=None):
        if middlewares is None:
            middlewares = []
        applied_middlewares = []
        ctrl_path, ctrl_middleware = self.get_middleware_of_handler(ctrl, path='path')
        # applied_middlewares += self.apply_middleware_to_path(module, ctrl_path, middlewares + ctrl_middleware)
        methods = inspect.getmembers(ctrl, predicate=Utils.is_handler)
        for name, value in methods:
            method_path, method_middleware = self.get_middleware_of_handler(value)
            applied_middlewares += self.apply_middleware_to_path(module, ctrl_path + method_path
                                                                 , middlewares + ctrl_middleware + method_middleware)
        return applied_middlewares

    def extract_middleware_of_module(self, module):
        module_middlewares: list[MiddlewareDict] = []
        middlewares: list = [p for p in getattr(module, 'providers') if hasattr(p, 'middleware__')]
        controllers = [ctrl for ctrl in module.controllers if hasattr(ctrl, 'controller__')]
        for ctrl in controllers:
            module_middlewares += self.apply_middleware_to_ctrl(module, ctrl, middlewares)
        setattr(module, 'middlewares__', module_middlewares)

        if hasattr(module, 'nestipy_module__') and getattr(module, 'nestipy_module__'):
            if hasattr(module, 'configure'):
                c = self.middleware_consumer
                module_instance = module()
                module_instance.configure(c)

    # COMPILE
    async def compile(self):
        await self.register_and_resolve_provider(self.module, init=True)
        return self.module



import os
import shutil

from minijinja import Environment


class TemplateGenerator:
    env: Environment

    def __init__(self):
        self.env = Environment(
            loader=self.loader
        )
        self.env.add_filter('capitalize', str.capitalize)
        self.env.add_filter('lower', str.lower)

    @classmethod
    def loader(cls, name):
        segments = []
        for segment in name.split("/"):
            if "\\" in segment or segment in (".", ".."):
                return None
            segments.append(segment)
        try:
            path = os.path.join(os.path.dirname(__file__), 'views', *segments)
            with open(path) as f:
                content = f.read()
                return content
        except (IOError, OSError):
            pass

    def render_template(self, template, **kwargs):
        return self.env.render_template(template, **kwargs)

    @classmethod
    def copy_project(cls, destination):
        shutil.copytree(os.path.join(os.path.dirname(__file__), 'project'), destination)



import asyncio
import inspect
import logging
import traceback
from typing import Any

from fastapi import FastAPI, Depends, APIRouter
from fastapi_utils.cbv import cbv
from fastapi_utils.inferring_router import InferringRouter

from nestipy.core.utils import Utils
from .platform import PlatformAdapter


class PlatformFastAPI(PlatformAdapter[FastAPI]):
    resolved: list = []
    module: Any
    router: APIRouter = APIRouter()

    async def setup(self, module):
        try:
            imports = module.imports if hasattr(module, 'imports') else []
            if module.__name__ not in self.resolved:
                await self.controller_to_fastapi_controller(module=module)
                self.resolved.append(module.__name__)
                for m in imports:
                    await self.setup(m)
        except Exception as e:
            tb = traceback.format_exc()
            logging.error(e)
            logging.error(tb)

    def create_server(self, *args, **kwargs) -> FastAPI:
        routes = self.get_handlers()
        self.app = FastAPI(*args, debug=True, **kwargs)
        for router in routes:
            self.app.include_router(router, include_in_schema=True)
        return self.app

    @classmethod
    def controller_method_to_handler(cls, router: APIRouter, method, prefix=''):
        assert hasattr(method, "path__"), f'Path not exist for {method}.'
        kwargs = method.kwargs__ if hasattr(method, "kwargs__") else {}
        path = f"{prefix.rstrip('/')}/{method.path__.lstrip('/')}"
        match str(method.method__.value).lower():
            case 'post':
                return router.post(path, **kwargs)(method)
            case 'put':
                return router.put(path, **kwargs)(method)
            case 'patch':
                return router.patch(path, **kwargs)(method)
            case 'delete':
                return router.delete(path, **kwargs)(method)
            case _:
                return router.get(path, **kwargs)(method)

    @staticmethod
    def module_providers_to_dependencies(module):
        providers = getattr(module, 'providers')
        dependencies = []
        for p in providers:
            dependencies.append(Depends(lambda: p))
        return dependencies

    async def controller_to_fastapi_controller(self, module):
        await asyncio.sleep(0.001)
        controllers: list = module.controllers
        wrapped_controller = []
        for ctrl in controllers:
            router = InferringRouter(
                tags=[str(ctrl.__name__).replace('Controller', '').capitalize()],
                dependencies=self.module_providers_to_dependencies(module)
            )
            if not hasattr(ctrl, 'controller__'):
                continue
            class_attrs = {}
            new_members = inspect.getmembers(ctrl)
            for name, value in new_members:
                if not name.startswith('__'):
                    class_attrs[name] = value
            methods = inspect.getmembers(ctrl, predicate=Utils.is_handler)
            for name, value in methods:
                class_attrs[name] = self.controller_method_to_handler(router=router, method=value,
                                                                      prefix=getattr(ctrl, 'path').rstrip('/'))
            attrs = {**class_attrs, "__module__": ctrl.__module__, '__name__': ctrl.__name__}
            new_controller: Any = type(ctrl.__name__, ctrl.__bases__, attrs)
            new_controller = cbv(router)(new_controller)
            wrapped_controller.append(new_controller)
            self.add_handler(router)
        module.controllers = wrapped_controller
        return module



from litestar.params import Body as LBody, Parameter as LParam, Dependency as LDepends
from fastapi.params import Body as FBody, Param as FParam, Depends as FDepends, Header as FHeader


class Body:
    def __new__(cls, *args, **kwargs):
        return cls



from dataclasses import dataclass, asdict
from typing import Optional, Callable, Awaitable

from masoniteorm.connections import ConnectionResolver

from nestipy.common import Module
from nestipy.core.module import NestipyModule
from nestipy.core.module.provider import ModuleProvider
from nestipy.plugins.dynamic_module.dynamic_module import DynamicModule

MASONITE_ORM_DB = 'MASONITE_ORM_DB'
MASONITE_ORM_CONFIG = 'MASONITE_ORM_CONFIG'


@dataclass
class MasoniteDatabaseConfig:
    driver: str
    host: str
    port: str
    user: str
    password: str
    database: str
    prefix: str = ''
    log_queries: bool = False


ConfigType = dict[str, MasoniteDatabaseConfig | str]


@Module(is_global=True)
class MasoniteOrmModule(DynamicModule, NestipyModule):
    @classmethod
    def for_root(cls, config: ConfigType):
        DB = ConnectionResolver()
        setattr(cls, 'DATABASES', config)
        setattr(cls, 'DB', DB)
        cls.register(provide=MASONITE_ORM_DB, value=DB)

    @classmethod
    def for_root_async(cls,
                       value: Optional[ConfigType] = None,
                       factory: Callable[..., Awaitable[ConfigType] | ConfigType] = None,
                       inject: list = None
                       ):
        if value is not None:
            DB = ConnectionResolver()
            setattr(cls, 'DATABASES', value)
            setattr(cls, 'DB', DB)
            return cls.register(provide=MASONITE_ORM_DB, value=DB)
        if factory is not None:
            return cls.register_async(
                provider=ModuleProvider(
                    provide=MASONITE_ORM_CONFIG,
                    use_factory=factory, inject=inject or []
                )
            )

    def on_startup(self):
        instances: dict = self.get_container().instances
        if MASONITE_ORM_CONFIG in instances.keys():
            config = MasoniteOrmModule.to_dict(instances[MASONITE_ORM_CONFIG])
            DB = ConnectionResolver().set_connection_details(config)
            setattr(self, 'DB', DB)
        else:
            DB: ConnectionResolver = getattr(self, 'DB', ConnectionResolver())
            DATABASES = getattr(self, 'DATABASES', {})
            DB.set_connection_details(MasoniteOrmModule.to_dict(DATABASES))
            setattr(self, 'DB', DB)

    @classmethod
    def get_db(cls):
        return getattr(cls, 'DB', None)

    @classmethod
    def to_dict(cls, config: dict[str, MasoniteDatabaseConfig]):
        config_dict = {}
        for key, value in config.items():
            config_dict[key] = asdict(value) if not isinstance(value, str) else value
        return config_dict



from typing import Callable

from nestipy.common.http_method import HttpMethod
from nestipy.core.utils import Utils


class Route:
    def __init__(self, method: HttpMethod, path='/', **kwargs):
        self.path = Utils.string_to_url_path(path)
        self.method = method
        self.kwargs = kwargs

    def __call__(self, func: Callable):
        setattr(func, "method__", self.method)
        setattr(func, "path__", self.path)
        setattr(func, "handler__", True)
        setattr(func, "kwargs__", self.kwargs)
        return func


class Get(Route):
    def __init__(self, path='/'):
        super().__init__(HttpMethod.GET, path)


class Post(Route):
    def __init__(self, path='/', **kwargs):
        super().__init__(HttpMethod.POST, path, **kwargs)


class Put(Route):
    def __init__(self, path='/', **kwargs):
        super().__init__(HttpMethod.PUT, path, **kwargs)


class Patch(Route):
    def __init__(self, path='/', **kwargs):
        super().__init__(HttpMethod.PATCH, path, **kwargs)


class Delete(Route):
    def __init__(self, path='/', **kwargs):
        super().__init__(HttpMethod.DELETE, path, **kwargs)



import inspect

from ...common.decorator.inject import Inject


class NestipyContainer:
    def __init__(self):
        self.dependencies = {}
        self.instances = {}

    def register(self, cls, token: str = None):
        if cls in self.dependencies:
            return
        if token is not None:
            self.dependencies[token] = cls
            return
        params = inspect.getmembers(cls, predicate=lambda member: isinstance(member, Inject))
        dependencies = {param[0]: param[1] for param in params}
        self.dependencies[cls] = dependencies

    def resolve_method(self, method, token):
        if token not in self.dependencies:
            self.register(method, token)
        if token in self.instances:
            return self.instances[token]
        instance = method()
        self.instances[token] = instance
        return instance

    def resolve_class_property_inject(self, cls, module, exist=None):
        dependency_resolved = []
        if exist is not None:
            dependency_resolved = []
        if cls not in self.dependencies:
            self.register(cls)
        if cls in self.instances:
            return self.instances[cls]

        providers_imports = self.get_module_exported_providers(module)
        deps = self.dependencies[cls].items()
        if len(deps) > 0:
            dependency_resolved.append(cls)
        for name, dep in deps:
            dependency_cls = dep.dependency
            if dependency_cls in dependency_resolved:
                raise Exception(f"Circular dependency found for {cls.__name__}.")

            if hasattr(module, 'provider_instances__') and dependency_cls in getattr(module,
                                                                                     'provider_instances__').keys():
                provider_instances__ = getattr(module, 'provider_instances__')
                dep_resolved = provider_instances__[dependency_cls]
                setattr(cls, name, dep_resolved)
            elif isinstance(dependency_cls, str):
                lambda_function = self.find_method_by_token(dependency_cls, providers_imports.keys())
                if lambda_function is not None:
                    dep_resolved = self.resolve_method(lambda_function, dependency_cls)
                    setattr(cls, name, dep_resolved)
                else:
                    setattr(cls, name, dep.default)
            else:
                if dependency_cls not in providers_imports.keys():
                    return setattr(cls, name, dep.default)
                dep_resolved = self.resolve(dependency_cls, providers_imports[dependency_cls], dependency_resolved)
                setattr(cls, name, dep_resolved)

    def resolve(self, cls, module, exist=None):
        self.resolve_class_property_inject(cls, module, exist)
        instance = cls()
        self.instances[cls] = instance
        return instance

    @classmethod
    def get_module_exported_providers(cls, module):
        providers = {}
        imports = [m for m in module.imports if m is not None]
        for m in imports:
            for p in m.exports:
                if hasattr(p, 'injectable__'):
                    providers[p] = m
        for p in module.providers:
            if hasattr(p, 'injectable__'):
                providers[p] = module
        return providers

    @classmethod
    def find_method_by_token(cls, token, imports):
        for p in imports:
            if hasattr(p, 'token__') and p.token__ == token:
                return p

        return None



import inspect
from snakecase import convert


def Injectable(_name: str = None):
    def wrapper(cls):
        setattr(cls, 'injectable__', True)
        setattr(cls, 'injectable__name__', convert(cls.__name__))
        if _name is not None:
            setattr(cls, 'token__', _name)
        return cls

    return wrapper



from nestipy.common.decorator import Injectable


@Injectable()
class AppService:

    @classmethod
    async def get(cls):
        return "test"

    @classmethod
    async def post(cls, data: str):
        return "test"

    @classmethod
    async def put(cls, id_: int, data: str):
        return "test"

    @classmethod
    async def delete(cls, id_: int):
        return "test"



from ...common.decorator.inject import Inject
from ...common.decorator.injectable import Injectable
from .constant import DATABASE_MODULE_TOKEN


@Injectable()
class DatabaseService:
    url: str = Inject(DATABASE_MODULE_TOKEN)

    def get_url(self, ) -> str:
        return self.url



from .platform import PlatformAdapter
from .platform_litestar import PlatformLitestar
from .platform_fastapi import PlatformFastAPI

__all__ = [
    'PlatformAdapter',
    'PlatformLitestar',
    'PlatformFastAPI'
]



from __future__ import annotations

import dataclasses
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    List,
    Mapping,
    Optional,
    Sequence,
    Type,
    Union,
)

from strawberry import BasePermission
from strawberry.annotation import StrawberryAnnotation
from strawberry.field import StrawberryField, _RESOLVER_TYPE
from strawberry.types.fields.resolver import StrawberryResolver

from .resolver import FieldResolver

if TYPE_CHECKING:
    from typing_extensions import Literal

    from strawberry.extensions.field_extension import FieldExtension


def field(
        resolver: Optional[_RESOLVER_TYPE[Any]] = None,
        *,
        name: Optional[str] = None,
        is_subscription: bool = False,
        description: Optional[str] = None,
        permission_classes: Optional[List[Type[BasePermission]]] = None,
        deprecation_reason: Optional[str] = None,
        default: Any = dataclasses.MISSING,
        default_factory: Union[Callable[..., object], object] = dataclasses.MISSING,
        metadata: Optional[Mapping[Any, Any]] = None,
        directives: Optional[Sequence[object]] = (),
        extensions: Optional[List[FieldExtension]] = None,
        graphql_type: Optional[Any] = None,
        init: Literal[True, False, None] = None,
) -> Any:
    type_annotation = StrawberryAnnotation.from_annotation(graphql_type)

    field_ = StrawberryField(
        python_name=None,
        graphql_name=name,
        type_annotation=type_annotation,
        description=description,
        is_subscription=is_subscription,
        permission_classes=permission_classes or [],
        deprecation_reason=deprecation_reason,
        default=default,
        default_factory=default_factory,
        metadata=metadata,
        directives=directives or (),
        extensions=extensions or [],
    )

    if resolver:
        assert init is not True, "Can't set init as True when passing a resolver."
        if not isinstance(resolver, StrawberryResolver):
            resolver = FieldResolver(resolver)
        return field_(resolver)
    return field_



from nestipy.common.decorator import Controller, Get, Inject, Post, Put, Delete
from .app_service import AppService


@Controller()
class AppController:
    service: AppService = Inject(AppService)

    @Get()
    async def get(self) -> str:
        return await self.service.get()

    @Post()
    async def post(self, data: str) -> str:
        return await self.service.post(data=data)

    @Put('/{user_id}')
    async def put(self, user_id: int, data: str) -> str:
        return await self.service.put(id_=user_id, data=data)

    @Delete('/{user_id}')
    async def delete(self, user_id: int) -> None:
        await self.service.delete(id_=user_id)



from abc import ABC, abstractmethod
from typing import Callable


class NestipyMiddleware(ABC):
    middleware__ = True

    @abstractmethod
    def use(self, scope, receive, send):
        return scope


class Middleware:
    middleware: list

    def __init__(self, *middleware):
        self.middleware = list(middleware)

    def __call__(self, cls):
        setattr(cls, 'middleware__', self.middleware)
        return cls



from dataclasses import asdict
from dataclasses import asdict
from typing import TYPE_CHECKING

from strawberry import Schema, type
from strawberry.asgi import GraphQL

from nestipy.common.decorator import NestipyMiddleware, Inject
from .constant import STRAWBERRY_MODULE_OPTION
from .override.field import field

if TYPE_CHECKING:
    from .strawberry_module import StrawberryOption


@type
class Query:
    @field
    def test(self) -> str:
        return 'Test ok'


class StrawberryMiddleware(NestipyMiddleware):
    option: 'StrawberryOption' = Inject(STRAWBERRY_MODULE_OPTION)

    def __init__(self):
        dict_option = asdict(self.option)
        if hasattr(self, 'schema'):
            self.app = GraphQL(schema=getattr(self, 'schema'), **dict_option)
        else:
            self.app = GraphQL(schema=Schema(query=Query), **dict_option)

    async def use(self, scope, receive, send):
        if hasattr(self, 'app'):
            await self.app(scope, receive, send)
            return 1



from nestipy.core.module.middleware import MiddlewareConsumer


class NestipyModule:
    nestipy_module__ = True

    def configure(self, consumer: MiddlewareConsumer):
        pass

    async def on_startup(self):
        pass

    async def on_shutdown(self):
        pass



class Module:
    def __init__(self, controllers=None, providers=None, imports=None, exports=None, is_global=False):
        self.controllers = controllers
        self.providers = providers
        self.imports = imports
        self.exports = exports
        self.is_global = is_global

    def __call__(self, cls):
        class_providers = []

        if hasattr(cls, 'providers'):
            class_providers = cls.providers
        class_attrs = {
            "controllers": self.controllers or [],
            "providers": class_providers + (self.providers or []) ,
            "imports": self.imports or [],
            "exports": self.exports or [],
            "is_module__": True,
            "is_global__": self.is_global,
        }
        for key, value in class_attrs.items():
            setattr(cls, key, value)
        return cls



import uvicorn

from .app_module import AppModule
from nestipy.core.factory import NestipyFactory

app = NestipyFactory.create(AppModule, title="My app")

if __name__ == '__main__':
    uvicorn.run('main:app', host="0.0.0.0", port=8000, reload=True)



from typing import Optional, Any

import click


class CliStyle:

    @classmethod
    def info(cls, info: Optional[Any] = None, ):
        click.secho(info, fg='green')

    @classmethod
    def error(cls, info: Optional[Any] = None, ):
        click.secho(info, fg='red')

    @classmethod
    def warning(cls, info: Optional[Any] = None, ):
        click.secho(info, fg='orange')

    @classmethod
    def success(cls, info: Optional[Any] = None, ):
        click.secho(info, fg='green', bold=True)



import asyncio
import inspect
import logging
import traceback

from litestar import Litestar, Controller, route
from litestar.di import Provide
from litestar.openapi import OpenAPIConfig

from .platform import PlatformAdapter
from ..utils import Utils


class PlatformLitestar(PlatformAdapter[Litestar]):
    resolved: list = []

    async def setup(self, module):
        try:
            imports = module.imports if hasattr(module, 'imports') else []
            if module.__name__ not in self.resolved:
                await self.controller_to_litestar_controller(module)
                self.resolved.append(module.__name__)
                for m in imports:
                    await self.setup(m)
        except Exception as e:
            tb = traceback.format_exc()
            logging.error(e)
            logging.error(tb)

    def create_server(self, title: str = "My App", version: str = "1.0.0", **kwargs) -> Litestar:
        handlers = self.get_handlers()
        self.app = Litestar(debug=True, route_handlers=handlers,
                            openapi_config=OpenAPIConfig(title, version, **kwargs))
        return self.app

    @staticmethod
    def controller_to_extend_litestar_controller(ctrl):
        class_attrs = {}
        for name, value in inspect.getmembers(ctrl):
            if not inspect.ismethod(value) and not inspect.isfunction(value) and not inspect.isbuiltin(
                    value) and name != '__slots__':
                class_attrs[name] = value

        if not hasattr(ctrl, 'tags'):
            class_attrs['tags'] = [str(ctrl.__name__).replace('Controller', '')]
        return type(ctrl.__name__, (Controller,),
                    {**class_attrs, "__module__": ctrl.__module__})

    @staticmethod
    def module_providers_to_dependencies(module):
        providers = getattr(module, 'providers')
        dependencies = {}
        for p in providers:
            dependencies[p.injectable__name__] = Provide(lambda: p, sync_to_thread=True)
        return dependencies

    async def controller_to_litestar_controller(self, module):
        await asyncio.sleep(0.001)
        controllers: list = module.controllers
        wrapped_controller = []
        for ctrl in controllers:
            ctrl.dependencies = self.module_providers_to_dependencies(module)
            if not hasattr(ctrl, 'controller__'):
                continue
            new_ctrl = self.controller_to_extend_litestar_controller(ctrl)
            class_attrs = {}
            new_members = inspect.getmembers(new_ctrl)
            for name, value in new_members:
                if not name.startswith('__'):
                    class_attrs[name] = value
            methods = inspect.getmembers(ctrl, predicate=Utils.is_handler)
            for name, value in methods:
                class_attrs[name] = self.controller_method_to_handler(value)
            modified_ctrl = type(new_ctrl.__name__, new_ctrl.__bases__,
                                 {**class_attrs, "__module__": new_ctrl.__module__})
            wrapped_controller.append(modified_ctrl)
            self.add_handler(modified_ctrl)
        module.controllers = wrapped_controller
        return module

    @classmethod
    def controller_method_to_handler(cls, method):
        assert hasattr(method, "path__"), f'Path not exist for {method}.'
        kwargs = method.kwargs__ if hasattr(method, "kwargs__") else {}
        sync_to_thread = True if not inspect.iscoroutinefunction(method) else None
        return route(path=method.path__, http_method=method.method__,
                     sync_to_thread=sync_to_thread, **kwargs)(method)



import logging
import traceback
from dataclasses import dataclass
from typing import Any, Literal, Optional

from nestipy.common.decorator import Module
from nestipy.core.module import NestipyModule
from nestipy.core.module.middleware import MiddlewareConsumer
from nestipy.plugins.dynamic_module.dynamic_module import DynamicModule
from nestipy.plugins.strawberry_module.compiler import GraphqlCompiler
from nestipy.plugins.strawberry_module.constant import STRAWBERRY_MODULE_OPTION
from nestipy.plugins.strawberry_module.strawberry_middleware import StrawberryMiddleware


@dataclass
class StrawberryOption:
    graphql_ide: Optional[Literal["graphiql", "apollo-sandbox", "pathfinder"]]


@Module(providers=[])
class StrawberryModule(DynamicModule, NestipyModule):
    resolvers = []

    @classmethod
    def for_root(cls, option: Any = StrawberryOption(graphql_ide='graphiql'), resolvers=None):
        if resolvers is None:
            resolvers = []
        setattr(cls, 'resolvers', resolvers)
        return cls.register(provide=STRAWBERRY_MODULE_OPTION, value=option)

    def configure(self, consumer: MiddlewareConsumer):
        if len(self.resolvers) > 0:
            try:
                gql_compiler = GraphqlCompiler(modules=self.resolvers)
                schema = gql_compiler.compile()
                setattr(StrawberryMiddleware, 'schema', schema)
                consumer.apply_for_route(self, '/graphql', StrawberryMiddleware)
            except Exception as e:
                tb = traceback.format_exc()
                logging.error(e)
                logging.error(tb)

    async def on_startup(self):
        pass



from strawberry import input, type as strawberry_type

from nestipy.common.decorator import Injectable


class Resolver:
    resolver__ = True

    def __call__(self, cls, *_):
        setattr(cls, 'graphql__resolver__', True)
        setattr(cls, 'type__', True)
        return Injectable()(cls)


class Query:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, cls):
        setattr(cls, 'graphql__resolver__', True)
        setattr(cls, 'query__', True)
        setattr(cls, 'kwargs__', self.kwargs)
        return cls


class Subscription:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, cls):
        setattr(cls, 'graphql__resolver__', True)
        setattr(cls, 'subscription__', True)
        setattr(cls, 'kwargs__', self.kwargs)
        return cls


class Mutation:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, cls):
        setattr(cls, 'graphql__resolver__', True)
        setattr(cls, 'mutation__', True)
        setattr(cls, 'kwargs__', self.kwargs)
        return cls


class Type:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, cls):
        setattr(cls, 'graphql__type__', True)
        return strawberry_type(cls, **self.kwargs)


class Input:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, cls):
        setattr(cls, 'graphql__input__', True)
        return input(cls, **self.kwargs)



from litestar.dto import DTOConfig, DataclassDTO
import inspect


def Dto():
    def wrapper(cls):
        class_attrs = {
            "config_module": DTOConfig(experimental_codegen_backend=True),
            "__module__": cls.__module__
        }
        for name, value in inspect.getmembers(cls):
            if not name.startswith('__'):
                class_attrs[name] = value

        return type(cls.__name__, (DataclassDTO[cls],) + cls.__bases__, {**class_attrs})

    return wrapper



import asyncio
import inspect
from typing import TypeVar, Generic

from ..common.enum.platform import PlatFormType
from ..core.app_context import AppNestipyContext
from ..core.platform import PlatformLitestar, PlatformAdapter

T = TypeVar('T', PlatformLitestar, PlatformAdapter, covariant=True)


class NestipyFactoryMeta(type):
    def __getitem__(self, item):
        setattr(self, '__generic_type__', item.__name__)
        return self


class NestipyFactory(metaclass=NestipyFactoryMeta):

    @classmethod
    def create(cls, module, **kwargs) -> AppNestipyContext:
        platform = PlatFormType.LITESTAR
        if hasattr(cls, '__generic_type__') and getattr(cls, '__generic_type__') == 'PlatformFastAPI':
            platform = PlatFormType.FASTAPI
        return AppNestipyContext(module=module, platform=platform, **kwargs)



from typing import Optional

from ...common.decorator.module import Module
from .config_service import ConfigService
from .constant import CONFIG_MODULE_TOKEN
from ..dynamic_module.dynamic_module import DynamicModule


@Module(providers=[ConfigService], exports=[ConfigService], is_global=True)
class ConfigModule(DynamicModule):

    @classmethod
    def for_root(cls, value: Optional = None):
        env_file = value or '.env'
        from dotenv import dotenv_values
        value = dotenv_values(env_file)
        return cls.register(value=value,provide=CONFIG_MODULE_TOKEN)



from nestipy.core.utils import Utils


def Controller(path: str = '/', **kwargs):
    def wrapper(cls):
        class_attrs = {
            "kwargs__": kwargs,
            "controller__": True,
            "path": Utils.string_to_url_path(path)
        }
        for key, value in class_attrs.items():
            setattr(cls, key, value)

        return cls

    return wrapper



import inspect
import re


class Utils:
    @classmethod
    def is_handler(cls, member):
        return inspect.isfunction(member) and hasattr(member, "handler__")

    @classmethod
    def string_to_url_path(cls, string):
        # Remove leading and trailing slashes for consistency.
        string = string.strip("/")
        # Replace invalid characters with underscores.
        valid_chars = r"[^a-zA-Z0-9\-_]"
        string = re.sub(valid_chars, '_', string)
        return '/' + string

    @classmethod
    def match_route(cls, route_pattern, url):
        pattern = re.compile(re.sub(r"{(\w+)}", r"(?P<\1>[^/]+)", route_pattern))
        match = pattern.match(url)
        return match is not None



from nestipy.common.decorator import Module
from .app_controller import AppController
from .app_service import AppService


@Module(
    controllers=[AppController],
    providers=[AppService]
)
class AppModule:
    pass



import click
import questionary

from nestipy.cli.handler import NestipyCliHandler
from nestipy.cli.style import CliStyle
from click_aliases import ClickAliasedGroup

handler = NestipyCliHandler()
echo = CliStyle()


@click.group(cls=ClickAliasedGroup)
def main():
    pass


@main.command()
@click.argument('name')
def new(name):
    """ Create new project """
    # if not shutil.which('poetry'):
    # curl -sSL https://install.python-poetry.org | python3 -
    click.clear()
    created = handler.create_project(name)
    if not created:
        echo.error(f"Folder {name} already exist.")
    echo.info(f"Project {name} created successfully.\nStart your project by running:\n\tcd {name}"
              f"\n\tpython -m pip install -r requirements.py\n\tpython main.py")
    # else:
    #     echo.error(f"Nestipy need poetry as dependency manager.")


@main.group(cls=ClickAliasedGroup, name='generate', aliases=['g', 'gen'])
def make():
    """ Generate resource, module, controller, service, resolver, graphql input """
    pass


@make.command(name='resource', aliases=['r', 'res'])
@click.argument('name')
def resource(name):
    """Create new resource for project."""
    name = str(name).lower()
    choice = questionary.select('Select resource type:', choices=['api', 'graphql']).ask()
    if choice == 'graphql':
        handler.generate_resource_graphql(name)
    else:
        handler.generate_resource_api(name)
    echo.success(f"Resource created successfully inside src/{name}.")


@make.command(aliases=['mod'])
@click.argument('name')
def module(name):
    """Create new module"""
    name = str(name).lower()
    handler.generate_module(name, prefix='single')
    echo.success(f"Module created successfully inside src/{name}.")


@make.command()
@click.argument('name')
def controller(name):
    """ Create new controller """
    name = str(name).lower()
    handler.generate_controller(name, prefix='single')
    echo.success(f"Controller created successfully inside src/{name}.")


@make.command()
@click.argument('name')
def resolver(name):
    """ Create new graphql resolver """
    handler.generate_resolver(name, prefix='single')
    echo.success(f"Resolver created successfully inside src/{name}.")


@make.command()
@click.argument('name')
def service(name):
    """ Create new service """
    name = str(name).lower()
    handler.generate_service(name, prefix='single')
    echo.success(f"Service created successfully inside src/{name}.")


@make.command(name='input')
@click.argument('name')
def graphql_input(name):
    """ Create new service """
    name = str(name).lower()
    handler.generate_service(name, prefix='single')
    echo.success(f"Graphql Input created successfully inside src/{name}.")


if __name__ == "__main__":
    main()



from ...common.decorator.module import Module
from .constant import DATABASE_MODULE_TOKEN
from .database_service import DatabaseService
from ..dynamic_module.dynamic_module import DynamicModule
from ...core.module.provider import ModuleProvider


@Module(providers=[DatabaseService], exports=[DatabaseService], is_global=True)
class DatabaseModule(DynamicModule):

    @classmethod
    def for_root_async(cls, value: ModuleProvider, inject=None):
        return cls.register_async(
            provider=ModuleProvider(
                use_value=value,
                provide=DATABASE_MODULE_TOKEN,
                inject=inject or []
            )
        )



from ...common.decorator.inject import Inject
from ...common.decorator.injectable import Injectable
from .constant import CONFIG_MODULE_TOKEN


@Injectable()
class ConfigService:
    env: dict = Inject(CONFIG_MODULE_TOKEN)

    def get(self, key, default=None) -> str:
        return self.env.get(key) or default



import os.path

from nestipy.common.templates.generator import TemplateGenerator


class NestipyCliHandler:
    generator: TemplateGenerator

    def __init__(self):
        self.generator = TemplateGenerator()

    def create_project(self, name) -> bool:
        destination = os.path.join(os.getcwd(), name)
        if os.path.exists(destination):
            return False
        self.generator.copy_project(destination)
        return True

    @classmethod
    def mkdir(cls, name):
        path = os.path.join(os.getcwd(), 'src', name)
        if not os.path.exists(path):
            os.mkdir(path)
            open(os.path.join(path, '__init__.py'), 'a').close()
        return path

    def generate_resource_api(self, name):
        self.generate_dto(name)
        self.generate_service(name)
        self.generate_controller(name)
        self.generate_module(name)

    def generate_resource_graphql(self, name):
        self.generate_input(name)
        self.generate_service(name)
        self.generate_resolver(name)
        self.generate_module(name, prefix='graphql')

    def generate_module(self, name: str, prefix: str = None):
        path = self.mkdir(name)
        self.generate(name, path, 'module', prefix=prefix)
        self.modify_app_module(name)

    def generate_controller(self, name: str, prefix: str = None):
        path = self.mkdir(name)
        self.generate(name, path, 'controller', prefix=prefix)

    def generate_service(self, name: str, prefix: str = None):
        path = self.mkdir(name)
        self.generate(name, path, 'service', prefix=prefix)

    def generate_resolver(self, name: str, prefix: str = None):
        path = self.mkdir(name)
        self.generate(name, path, 'resolver', prefix=prefix)

    def generate_dto(self, name: str, prefix: str = None):
        path = self.mkdir(name)
        self.generate(name, path, 'dto', prefix=prefix)

    def generate_input(self, name: str, prefix: str = None):
        path = self.mkdir(name)
        self.generate(name, path, 'input', prefix=prefix)

    def generate(self, name, parent_path, template, prefix: str = None):
        content = self.generator.render_template(
            f'{prefix}_' if prefix is not None else ''
                                                    f"{template}.txt", name=name)
        file_path = str(os.path.join(parent_path, f"{name.lower()}_{template}.py"))
        print(file_path)
        f = open(file_path, 'w+')
        f.write(content)
        f.close()

    @classmethod
    def modify_app_module(cls, name):
        name = str(name)
        app_path = os.path.join(os.getcwd(), 'src', 'app_module.py')
        if os.path.exists(app_path):
            new_import = f"{str(name).capitalize()}Module"
            with open(app_path, 'r') as file:
                file_content = file.read()
                file.close()
                module_pattern = r'@Module\(([^)]*)\)'
                text_to_add = f'from .{name.lower()}.{name.lower()}_module import {name.capitalize()}Module'
                import re
                match = re.search(module_pattern, file_content)
                if match:
                    existing_imports_str = match.group(1)
                    existing_imports_match = re.search(r'imports=\[(.*?)\]', existing_imports_str)
                    if existing_imports_match:
                        existing_imports = existing_imports_match.group(1)
                        new_imports = existing_imports + ', ' + new_import
                        modified_imports_str = re.sub(r'imports=\[(.*?)\]', 'imports=[' + new_imports + ']',
                                                      existing_imports_str)
                        modified_content = file_content.replace(match.group(0),
                                                                text_to_add + '\n@Module(' + modified_imports_str + ')')
                    else:
                        # If imports=[] doesn't exist, add imports directly
                        modified_content = file_content.replace(match.group(0),
                                                                text_to_add + f'\n@Module(\n\timports=[{new_import}],'
                                                                              f'{existing_imports_str})')
                    with open(app_path, 'w') as file2:
                        file2.write(modified_content)
                        file2.close()
                else:
                    print("No @Module decorator found in the file.")



from dataclasses import dataclass
from typing import Optional, Any, Callable


@dataclass
class ModuleProvider:
    provide: Optional[Any] = None
    use_value: Optional[Any] = None
    use_factory: Optional[Callable[[Any], Any]] = None
    inject: Optional[list[Any]] = None



from abc import abstractmethod, ABC
from typing import Generic, TypeVar, Callable, Any

T = TypeVar('T')


class PlatformAdapter(Generic[T], ABC):
    app: T
    _handlers: list[Callable] = []
    _hooks: list[Callable] = []
    _listeners: list[Callable] = []

    def get_handlers(self):
        return self._handlers

    def add_handler(self, handler: Callable):
        self._handlers.append(handler)

    def get_hooks(self):
        return self._hooks

    def listeners(self):
        return self._listeners

    def add_listener(self, listener: Callable):
        self._listeners.append(listener)

    @abstractmethod
    async def setup(self, module):
        pass

    @abstractmethod
    def create_server(self, *args, **kwargs) -> T:
        pass



import inspect
from abc import ABC
from typing import TypeVar

from nestipy.core.module.provider import ModuleProvider

T = TypeVar('T')


class DynamicModule(ABC):
    providers = []
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls, *args, **kwargs)
        return cls._instance

    def get_container(self):
        if hasattr(self, 'container__'):
            return getattr(self, 'container__')
        else:
            return None

    @classmethod
    def _create_module(cls, provider: ModuleProvider = None):
        members = inspect.getmembers(cls, predicate=lambda a: not inspect.ismethod(a))
        class_attrs = {}
        if provider is not None:
            class_attrs['module_provider__'] = provider
        for name, value in members:
            if not name.startswith('__'):
                class_attrs[name] = value
        return type(cls.__name__, (DynamicModule,), class_attrs)

    @classmethod
    def register(cls, provide=None, value=None, inject: list = None):
        provider = None
        if provide is not None and value is not None:
            provider = ModuleProvider(provide=provide, use_value=value, inject=inject or [])
        return cls._create_module(provider=provider)

    @classmethod
    def register_async(cls, provider: ModuleProvider = None):
        return cls._create_module(provider=provider)



from .app_context import AppNestipyContext
from .factory import NestipyFactory
from .ioc.container import NestipyContainer
from .module.compiler import ModuleCompiler
from .platform.platform import PlatformAdapter
from .platform.platform_litestar import PlatformLitestar
from .platform.platform_fastapi import PlatformFastAPI

__all__ = [
    'NestipyFactory',
    'AppNestipyContext',
    'NestipyContainer',
    'ModuleCompiler',
    'PlatformAdapter',
    'PlatformLitestar',
    'PlatformFastAPI'
]



from .controller import Controller
from .inject import Inject
from .module import Module
from .methods import Get, Post, Route, Put, Patch, Delete
from .dto import Dto
from .injectable import Injectable
from .middleware import NestipyMiddleware

__all__ = [
    'Controller',
    'Inject',
    'Module',
    'Get',
    'Post',
    'Route',
    'Put',
    'Patch',
    'Delete',
    'Dto',
    'NestipyMiddleware',
    'Injectable'
]



from typing import Optional, Any


class Inject:
    def __init__(self, cls, default: Optional = None):
        if not isinstance(cls, str) and not hasattr(cls, "injectable__"):
            raise Exception(f"{cls.__name__} is not injectable")
        setattr(self, 'default', default)
        setattr(self, 'dependency', cls)



class Catch:
    def __init__(self, handler):
        self.handler = handler

    def __call__(self, cls):
        error_handlers = getattr(cls, 'error_handler__', [])
        error_handlers = error_handlers + (self.handler if isinstance(self.handler, list) else [self.handler])
        setattr(cls, 'error_handler__', error_handlers)
        return cls



import inspect
import json
import logging
import traceback
from dataclasses import dataclass, asdict
from typing import Any, Literal, Callable, Awaitable
from peewee import MySQLDatabase, PostgresqlDatabase, SqliteDatabase, Database, Model as BaseModel

from .constant import PEEWEE_MODULE_CONFIG
from .peewee_service import PeeweeService
from ..dynamic_module.dynamic_module import DynamicModule
from ...common.decorator.module import Module
from ...core.module.provider import ModuleProvider


@dataclass
class PeeweeDatabaseConfig:
    driver: Literal['mysql', 'postgres', 'sqlite']
    database: str
    host: str
    port: int
    user: str
    password: str


@Module(providers=[PeeweeService], exports=[PeeweeService], is_global=True)
class PeeweeModule(DynamicModule):
    models = []
    db: Database

    async def on_startup(self):
        container_instance = self.get_container().instances
        if PEEWEE_MODULE_CONFIG in container_instance.keys():
            config: PeeweeDatabaseConfig = container_instance[PEEWEE_MODULE_CONFIG]
            self.create_db(config)
            models = []
            for m in self.models:
                new_m = PeeweeModule.class_to_model(m, self.db)
                models.append(new_m)
            if not self.db.is_closed():
                self.db.close()
            self.db.connect()
            self.db.create_tables(self.models)

    async def on_shutdown(self):
        if self.db is not None:
            if not self.db.is_closed():
                self.db.close()

    def create_db(self, config: PeeweeDatabaseConfig):
        config_dict = asdict(config)
        del config_dict['driver']
        database = config_dict['database']
        del config_dict['database']
        match config.driver:
            case 'postgres':
                self.db = PostgresqlDatabase(database, **config_dict)
            case 'sqlite':
                self.db = SqliteDatabase(database, **config_dict)
            case _:
                self.db = MySQLDatabase(database, **config_dict)

    @classmethod
    def for_root_async(cls,
                       value: PeeweeDatabaseConfig = None,
                       use_factory: Callable[[...], Awaitable[PeeweeDatabaseConfig] | PeeweeDatabaseConfig] = None,
                       inject: list = None
                       ):
        return super().register_async(
            provider=ModuleProvider(
                use_value=value,
                provide=PEEWEE_MODULE_CONFIG,
                use_factory=use_factory,
                inject=inject or [])
        )

    @classmethod
    def for_feature(cls, models: list):
        peewee_models = []
        for model in models:
            if hasattr(model, 'peewee_model__'):
                peewee_models.append(model)
        cls.models = cls.models + peewee_models
        return None

    @staticmethod
    def to_json(self):
        r = {}
        members = inspect.getmembers(self,
                                     predicate=lambda a: not inspect.ismethod(a) and not inspect.isfunction(a))
        excludes = ['DoesNotExist', 'dirty_fields']
        filtered_members = [me for me in members if
                            not me[0].startswith('_') and not me[0].endswith('__') and me[0] not in excludes]
        for k, v in filtered_members:
            try:
                r[k] = v
            except Exception as e:
                tb = traceback.format_exc()
                logging.error(e)
                logging.error(tb)
                r[k] = json.dumps(getattr(self, k))
        return r

    @staticmethod
    def class_to_model(m, db):
        class_attrs = {"__module__": m.__module__, 'to_json': PeeweeModule.to_json}
        for name, value in inspect.getmembers(m):
            if not name.startswith('__'):
                class_attrs[name] = value
        meta = type('Meta', (), {'database': db, 'table_name': str(m.__name__).lower()})
        setattr(m, "Meta", meta)
        class_attrs['Meta'] = meta
        new_cls = type(m.__name__, (BaseModel,) + m.__bases__, class_attrs)
        new_class_member = inspect.getmembers(new_cls)
        for name, value in new_class_member:
            if not name.startswith('__'):
                setattr(m, name, value)

        globals()[m.__name__] = new_cls
        return new_cls



class MiddlewareConsumer:

    def __init__(self, compiler):
        self.compiler = compiler

    def apply_for_route(self, module, path, middleware):
        self.compiler.apply_middleware_to_path(module, path, [middleware])

    def apply_for_controller(self, module, ctrl, middleware):
        self.compiler.apply_middleware_to_ctrl(module, ctrl, [middleware])



from ...common.decorator.inject import Inject
from ...common.decorator.injectable import Injectable
from .constant import PEEWEE_MODULE_CONFIG


@Injectable()
class PeeweeService:
    config: dict = Inject(PEEWEE_MODULE_CONFIG)

    def get_config(self, ) -> dict:
        return self.config



import inspect

import snakecase
from strawberry import type as strawberry_type, Schema
from strawberry.tools import merge_types

from .override.field import field


class GraphqlCompiler:
    def __init__(self, modules):
        self.modules = modules

    def compile(self, **kwargs) -> Schema:
        queries, mutations, subscriptions = self.split_class_query_mutation()
        ComboQuery = merge_types("Query", tuple(queries)) if len(queries) > 0 else None
        ComboMutation = merge_types("Mutation", tuple(mutations)) if len(mutations) > 0 else None
        ComboSubscription = merge_types("Subscription", tuple(subscriptions)) if len(subscriptions) > 0 else None
        return Schema(query=ComboQuery, mutation=ComboMutation, subscription=ComboSubscription, **kwargs)

    def get_resolver(self):
        key = 'provider_instances__'
        resolvers = {}
        for m in self.modules:
            module_resolvers_instances = getattr(m, key) if hasattr(m, key) else []
            for key, value in module_resolvers_instances.items():
                if hasattr(value, 'graphql__resolver__'):
                    resolvers[key] = value
        return resolvers

    def split_class_query_mutation(self):
        class_query = []
        class_mutation = []
        class_subscription = []
        resolvers = self.get_resolver()
        for class_, instance_ in resolvers.items():
            name: str = self.get_name(class_)
            properties = self.extract_class_property(class_, instance_)
            class_methods = self.extract_class_not_resolver(class_)
            instance = self.create_instance_of_class(class_.__name__, properties + class_methods)
            methods = self.put_metadata_to_resolver(instance, self.extract_class_method_resolver(class_))

            query_resolver = self.extract_class_query(methods)
            query_ = strawberry_type(type(name + 'Query', (), {'__module__': class_.__module__,
                                                               **self.tuple_to_dict(query_resolver)}))
            class_query.append(query_)

            mutation_resolver = self.extract_class_mutation(methods)
            if len(mutation_resolver) > 0:
                mutation_ = strawberry_type(type(name + 'Mutation', (), {'__module__': class_.__module__,
                                                                         **self.tuple_to_dict(mutation_resolver)}))
                class_mutation.append(mutation_)

            subscription_resolver = self.extract_class_subscription(methods)
            if len(subscription_resolver) > 0:
                subscription_ = strawberry_type(type(name + 'Subscription', (), {'__module__': class_.__module__,
                                                                                 **self.tuple_to_dict(
                                                                                     subscription_resolver)}))
                class_subscription.append(subscription_)

        return class_query, class_mutation, class_subscription

    @classmethod
    def put_metadata_to_resolver(cls, instance, resolvers: list[tuple]):
        modified_resolvers: list[tuple] = []
        for key, method in resolvers:
            setattr(method, 'metadata__', instance)
            modified_resolvers.append((key, method))
        return modified_resolvers

    def create_instance_of_class(self, name, properties):
        cls = type(name, (), {**self.tuple_to_dict(properties)})
        return cls()

    @classmethod
    def extract_class_property(cls, resolver, instance_):
        properties = inspect.getmembers(resolver,
                                        predicate=lambda p: not inspect.ismethod(p) and not inspect.isfunction(p))
        return [(p[0], getattr(instance_, p[0])) for p in properties if not p[0].startswith('_')
                and not p[0].endswith('_')]

    @classmethod
    def is_resolver(cls, p):
        return hasattr(p, 'graphql__resolver__')

    @classmethod
    def extract_class_method_resolver(cls, resolver):
        methods = inspect.getmembers(resolver,
                                     predicate=lambda p:
                                     (inspect.ismethod(p) or inspect.isfunction(p)) and cls.is_resolver(p))

        return [p for p in methods if not p[0].startswith('__')]

    @classmethod
    def extract_class_not_resolver(cls, resolver):
        methods = inspect.getmembers(resolver,
                                     predicate=lambda p:
                                     (inspect.ismethod(p) or inspect.isfunction(p))
                                     and not cls.is_resolver(p))
        return [(m[0], m[1]) for m in methods if not m[0].startswith('__')]

    @classmethod
    def extract_class_query(cls, methods):
        return [(m[0], field(m[1], **cls.get_kwargs(m[1]))) for m in methods if hasattr(m[1], 'query__')]
        pass

    @classmethod
    def extract_class_mutation(cls, methods):
        return [(m[0], field(m[1], **cls.get_kwargs(m[1]))) for m in methods if hasattr(m[1], 'mutation__')]

    @classmethod
    def extract_class_subscription(cls, methods):
        return [(m[0], field(m[1], **cls.get_kwargs(m[1]), is_subscription=True)) for m in methods if
                hasattr(m[1], 'subscription__')]

    @classmethod
    def tuple_to_dict(cls, tuples: list):
        dicts = {}
        for key, value in tuples:
            dicts[key] = value
        return dicts

    @classmethod
    def get_name(cls, class_):
        return snakecase.convert(class_.__name__).split('_')[0].capitalize()

    @classmethod
    def get_kwargs(cls, method):
        return method.kwargs if hasattr(method, 'kwargs') else {}
